package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.es.model.ArtemisExerciseInfo;
import at.ac.uibk.gitsearch.properties.ApplicationProperties;
import at.ac.uibk.gitsearch.service.dto.MetadataUserDTO;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.json.JsonMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.fasterxml.jackson.dataformat.yaml.YAMLGenerator;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.WebTarget;
import net.minidev.json.JSONObject;
import net.minidev.json.JSONStyle;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.eclipse.jgit.lib.Repository;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.Group;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

@Service
public class ZipService {

    private static final String TMP_DIR_PROPERTY = "java.io.tmpdir";

    private final Logger log = LoggerFactory.getLogger(ExerciseImportService.class);

    private final FileService fileService;

    public ZipService(FileService fileService) {
        this.fileService = fileService;
    }

    /**
     * Used to temporary store content of Artemis' exercises Zip-Files into a
     * tmp-subdirectory
     *
     * Recursion is used in case of nested Zip-Files
     *
     * @param zipInputStream containing the Zip-File to unzip
     * @param subDir         name of the tmp-subdirectory
     */
    @SuppressWarnings({ "PMD.AssignmentInOperand", "PMD.CognitiveComplexity" })
    public void storeZipInTmp(ZipInputStream zipInputStream, String subDir) throws IOException {
        Path targetDir;

        if (subDir.startsWith(System.getProperty(TMP_DIR_PROPERTY))) {
            targetDir = Files.createDirectories(Paths.get(subDir));
        } else {
            targetDir = Files.createDirectories(Paths.get(System.getProperty(TMP_DIR_PROPERTY), subDir));
        }

        ZipEntry zipEntry;

        while ((zipEntry = zipInputStream.getNextEntry()) != null) {
            if (zipEntry.getName().contains(".git")) { // important as Artemis exercises come in submodules
                continue;
            }

            Path zipEntryPath = Paths.get(targetDir.toString(), zipEntry.getName());

            if (isZipFile(zipEntry)) {
                writeFile(zipEntryPath.toFile(), zipInputStream);
                String newSubDir;
                if (StringUtils.endsWithAny(zipEntry.getName(), new String[] { "-exercise.zip", "-solution.zip", "-tests.zip" })) {
                    newSubDir = Paths.get(targetDir.getParent().toString(), getRepoNameFromZipEntry(zipEntry)).toString();
                } else {
                    newSubDir =
                        Paths.get(targetDir.toString(), zipEntry.getName().substring(0, zipEntry.getName().lastIndexOf("."))).toString();
                }

                storeZipInTmp(new ZipInputStream(new FileInputStream(zipEntryPath.toFile())), newSubDir);
                if (Files.deleteIfExists(zipEntryPath)) {
                    log.warn("Unable to delete temporary Zip file [{}] after extraction. Skipping..", zipEntryPath);
                }
            } else if (zipEntry.isDirectory()) {
                Files.createDirectory(zipEntryPath);
            } else {
                writeFile(zipEntryPath.toFile(), zipInputStream);
            }
        }
        cleanUpTmpDirectory(targetDir.toFile());
        zipInputStream.close();
    }

    /**
     * Utility method used to check if a ZipEntry represents a Zip file
     *
     * @param zipEntry to check
     * @return true if zipEntry is a Zip file, false otherwise
     */
    private boolean isZipFile(ZipEntry zipEntry) {
        return zipEntry.getName().endsWith(".zip");
    }

    /**
     * Cleans up a temporary directory by removing it if empty
     * or scheduling its deletion with a delay of 60 minutes, otherwise.
     *
     * @param tmpDir to clean up
     */
    private void cleanUpTmpDirectory(File tmpDir) {
        try {
            if (tmpDir.exists() && tmpDir.isDirectory() && fileService.isEmpty(tmpDir.toPath())) {
                Files.delete(tmpDir.toPath());
            }
            fileService.scheduleForDeletion(tmpDir.toPath(), 60);
        } catch (IOException e) {
            log.error("An error occurred while cleaning up a temporary directory: ", e);
        }
    }

    /**
     * Utility method used to extract the repository name from Zip files
     * representing git submodules. (Artemis exercise exports come in
     * submodules for 'exercise', 'solution' and 'test')
     *
     * @param zipEntry to get the repository name from
     * @return the repository name
     */
    private String getRepoNameFromZipEntry(ZipEntry zipEntry) {
        return zipEntry.getName().substring(zipEntry.getName().lastIndexOf("-") + 1, zipEntry.getName().lastIndexOf("."));
    }

    /**
     * Method used to write files from a zipInputStream
     *
     * @param entryFile      File to write
     * @param zipInputStream Byte stream to read from
     * @throws IOException
     */
    @SuppressWarnings("PMD.AssignmentInOperand")
    private void writeFile(File entryFile, ZipInputStream zipInputStream) throws IOException {
        byte[] buffer = new byte[1024];

        if (!entryFile.getParentFile().exists()) {
            Files.createDirectories(entryFile.getParentFile().toPath());
        }

        try (FileOutputStream fos = new FileOutputStream(entryFile)) {
            int len;
            while ((len = zipInputStream.read(buffer)) > 0) {
                fos.write(buffer, 0, len);
            }
        }
    }
}
