import { HttpClient, HttpHandler } from '@angular/common/http';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { ActivatedRoute } from '@angular/router';
import { ExerciseImportComponent } from 'app/exercise/import/exercise-import.component';
import { ExerciseService } from 'app/exercise/service/exercise.service';
import { ArtemisExerciseInfo } from 'app/shared/model/artemis-exercise-info.model';
import { IInteractivityType } from 'app/shared/model/exercise.model';
import { GitlabGroup, IVisibility } from 'app/shared/model/gitlab-group.model';
import { SearchResultDTO } from 'app/shared/model/search/search-result-dto.model';
import { of, throwError } from 'rxjs';
import { TranslateModule } from '@ngx-translate/core';

describe('Exercise Import Component', () => {
  let comp: ExerciseImportComponent;
  let fixture: ComponentFixture<ExerciseImportComponent>;
  let service: ExerciseService;

  const exerciseInfoTemplate = new ArtemisExerciseInfo(
    [],
    '',
    [],
    [],
    false,
    '',
    0,
    '',
    '',
    [],
    [],
    [],
    '',
    '',
    '',
    {} as IInteractivityType,
    [],
    [],
    [],
    new Date(),
    '',
    '',
    '',
    {} as SearchResultDTO,
    ['Java'],
    '',
    [],
    0,
    [],
    '',
    '',
    [],
    [],
    '',
    'testExTitle',
    '',
    '',
    0,
    0,
    0,
    false
  );

  const group: GitlabGroup = {
    id: 1,
    name: 'testGroup',
    path: 'testPath',
    description: '',
    visibility: IVisibility.PUBLIC,
    avatarUrl: '',
    webUrl: '',
    fullName: '',
    fullPath: '',
    parentId: 0,
  };

  const newPerson = { name: 'newPerson', email: 'person@email', affiliation: 'newAffiliation' };
  const newPersonCopy = { name: 'newPerson', email: 'person@email', affiliation: 'newAffiliation' };
  const personToRemove = { name: 'personToRemove', email: 'person@toremove', affiliation: '' };

  beforeEach(() => {
    TestBed.configureTestingModule({
      imports: [TranslateModule.forRoot()],
      declarations: [ExerciseImportComponent],
      providers: [
        {
          provide: ActivatedRoute,
          useValue: {
            snapshot: {
              params: {
                token: 'exercise-test-token',
              },
              queryParams: {
                callback: 'https://callback-url.test',
              },
            },
          },
        },
        ExerciseImportComponent,
        ExerciseService,
        HttpClient,
        HttpHandler,
      ],
    })
      .overrideTemplate(ExerciseImportComponent, '')
      .compileComponents();

    fixture = TestBed.createComponent(ExerciseImportComponent);
    comp = fixture.componentInstance;
    service = TestBed.inject(ExerciseService);
  });

  describe('init', () => {
    it('Should load on init', () => {
      // GIVEN
      jest.spyOn(service, 'importExerciseInfoFromArtemis').mockImplementation(() => of(exerciseInfoTemplate));

      // WHEN
      comp.ngOnInit();

      // THEN
      expect(comp.exerciseInfo?.title).toEqual('testExTitle');
      expect(comp.exerciseInfo?.programmingLanguage[0]).toEqual('Java');
    });
  });

  describe('map metadata', () => {
    it('Should map metadata values when importing', () => {
      // GIVEN
      comp.exerciseToken = 'TestToken';
      comp.exerciseInfo = exerciseInfoTemplate;
      comp.exerciseInfo.difficulty = 'ADVANCED';
      comp.isPublisherSameAsCreator = true;

      jest.spyOn(comp, 'openGitlabPathSelectorModal').mockImplementation(
        () =>
          new Promise(resolve => {
            resolve(group);
          })
      );
      jest.spyOn(service, 'submitExerciseInfoForImport').mockImplementation(() => of({} as ArtemisExerciseInfo));

      // WHEN
      comp.import();

      // THEN
      expect(comp.exerciseInfo.difficulty).toEqual(comp.exerciseInfo.difficulty.toLowerCase());
      expect(comp.exerciseInfo.publisher).toEqual(comp.exerciseInfo.creator);
    });
  });

  describe('add and remove creators', () => {
    it('Should add and remove creators from list', () => {
      // GIVEN
      comp.exerciseInfo = exerciseInfoTemplate;

      // WHEN
      comp.addPerson(newPerson, 'creator');
      comp.addPerson(personToRemove, 'creator');
      comp.removePerson(personToRemove, 'creator');

      // THEN
      expect(comp.exerciseInfo?.creator).toEqual([newPersonCopy]);
    });
  });

  describe('add and remove publishers', () => {
    it('Should add and remove publishers from list', () => {
      // GIVEN
      comp.exerciseInfo = exerciseInfoTemplate;

      // WHEN
      comp.addPerson(newPerson, 'publisher');
      comp.addPerson(personToRemove, 'publisher');
      comp.removePerson(personToRemove, 'publisher');

      // THEN
      expect(comp.exerciseInfo?.publisher).toEqual([newPersonCopy]);
    });
  });

  describe('validate button groups', () => {
    it('Should validate form button groups', () => {
      // GIVEN
      comp.exerciseInfo = exerciseInfoTemplate;
      comp.exerciseInfo.keyword = [];
      comp.exerciseInfo.language = [];
      const resultWithErrors = comp.hasValidationErrors();

      comp.exerciseInfo.keyword.push('keyword');
      comp.setLanguage('english');
      const resultWithoutErrors = comp.hasValidationErrors();

      // THEN
      expect(resultWithErrors).toEqual(true);
      expect(resultWithoutErrors).toEqual(false);
    });
  });

  describe('display import errors', () => {
    it('should set unauthorized error', () => {
      // GIVEN
      const error = {
        status: 401,
        message: 'Unauthorized',
      };

      jest.spyOn(service, 'importExerciseInfoFromArtemis').mockReturnValue(throwError(() => error));

      // WHEN
      comp.ngOnInit();

      // THEN
      expect(comp.errorStatus).toBeDefined();
      expect(comp.occurredError).toEqual('exercise.import.error.unauthorized');
    });

    it('should set not found error', () => {
      // GIVEN
      const error = {
        status: 404,
        message: 'Not Found',
      };

      jest.spyOn(service, 'importExerciseInfoFromArtemis').mockReturnValue(throwError(() => error));

      // WHEN
      comp.ngOnInit();

      // THEN
      expect(comp.errorStatus).toBeDefined();
      expect(comp.occurredError).toEqual('exercise.import.error.notFound');
    });

    it('should set error', () => {
      // GIVEN
      const error = {
        status: 500,
        message: 'Internal Server Error',
      };

      jest.spyOn(service, 'importExerciseInfoFromArtemis').mockReturnValue(throwError(() => error));

      // WHEN
      comp.ngOnInit();

      // THEN
      expect(comp.errorStatus).toBeDefined();
      expect(comp.occurredError).toEqual('exercise.import.error.loadingError');
    });
  });

  describe('Validate metadata on import', () => {
    it('Should validate metadata on import', () => {
      // GIVEN
      comp.exerciseInfo = exerciseInfoTemplate;
      comp.exerciseInfo.keyword = [];
      comp.exerciseInfo.language = [];
      jest.spyOn(service, 'validateExerciseInfo').mockReturnValue(of('0'));

      // WHEN
      comp.import();

      // THEN
      expect(service.validateExerciseInfo).toHaveBeenCalled();
    });
  });
});
